/* 
 * <LIC_AMD_STD>
 * Copyright <2000-2010> ADVANCED MICRO DEVICES, INC. All Rights Reserved.
 *    
 *    
 * Unless otherwise designated in writing, this software and any related 
 * documentation are the confidential proprietary information of AMD. 
 *   
 * THESE MATERIALS ARE PROVIDED "AS IS" WITHOUT ANY UNLESS OTHERWISE 
 * NOTED IN WRITING, EXPRESS OR IMPLIED WARRANTY OF ANY KIND, INCLUDING 
 * BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, NONINFRINGEMENT, 
 * TITLE, FITNESS FOR ANY PARTICULAR PURPOSE AND IN NO EVENT SHALL AMD 
 * OR ITS LICENSORS BE LIABLE FOR ANY DAMAGES WHATSOEVER. 
 *   
 * AMD does not assume any responsibility for any errors which may appear 
 * in the Materials nor any responsibility to support or update the Materials.  
 * AMD retains the right to modify the Materials at any time, without notice, 
 * and is not obligated to provide such modified Materials to you. AMD is 
 * not obligated to furnish, support, or make any further information 
 * available to you.
 * 
 * </LIC_AMD_STD>  
 * <CTL_AMD_STD>
 * </CTL_AMD_STD>  
 * <DOC_AMD_STD>
 * </DOC_AMD_STD>  
 */

// Every debug output has "Modulname text"
static char STR_MODULENAME[] = "GEODE Stream: ";

#include "duraudio.h"

#include "minwave.h"
#include "GEODEwave.h"

extern PDURAUDIO pGeode;



/*****************************************************************************
 * General Info
 *****************************************************************************
 * To protect the stPRDList structure that is used to store mappings, we use a
 * spin lock called MapLock. This spin lock is also acquired when we change
 * the DMA registers. Normally, changes in stPRDList and the DMA registers go
 * hand in hand. In case we only want to change the DMA registers, we need
 * to acquire the spin lock!
 */

#pragma code_seg("PAGE")
/*****************************************************************************
 * CreateMiniportWaveGEODEStream
 *****************************************************************************
 * Creates a wave miniport stream object for the GEODE audio adapter. This is
 * (nearly) like the macro STD_CREATE_BODY_ from STDUNK.H.
 */
NTSTATUS CreateMiniportWaveGEODEStream
(
    OUT CMiniportWaveGEODEStream  **WaveGeodeStream,
    IN  PUNKNOWN                pUnknownOuter,
    IN  POOL_TYPE               PoolType
)
{
    PAGED_CODE ();

    DOUT (DBG_FUNC, ("[CreateMiniportWaveGEODEStream]"));

    //
    // This is basically like the macro at stdunk with the change that we
    // don't cast to interface unknown but to interface WaveGeodeStream.
    //
    *WaveGeodeStream = new (PoolType, 'rCcP')
                        CMiniportWaveGEODEStream (pUnknownOuter);
    if (*WaveGeodeStream)
    {
        (*WaveGeodeStream)->AddRef ();
        return STATUS_SUCCESS;
    }

    return STATUS_INSUFFICIENT_RESOURCES;
}
   

/*****************************************************************************
 * CMiniportWaveGEODEStream::~CMiniportWaveGEODEStream
 *****************************************************************************
 * Destructor
 */
CMiniportWaveGEODEStream::~CMiniportWaveGEODEStream ()
{
    PAGED_CODE ();


    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::~CMiniportWaveGEODEStream]"));

    //
    // Print information about the scatter gather list.
    //
    DOUT (DBG_DMA, ("Head %d, Tail %d, Tag counter %d, Entries %d.",
                   stPRDList.nHead, stPRDList.nTail, stPRDList.ulTagCounter, 
                   stPRDList.nPRDEntries));

    if (Wave)
    {
        //
        // Disable interrupts and stop DMA just in case.
        //
        if (Wave->AdapterCommon)
        {
			ResetDMA();

            //
            // Update also the topology miniport if this was the render stream.
            //
            if (Wave->AdapterCommon->GetMiniportTopology () &&
                (Channel == PIN_WAVEOUT_OFFSET))
            {
                Wave->AdapterCommon->GetMiniportTopology ()->SetCopyProtectFlag (FALSE);
            }
        }
        
        //
        // Remove stream from miniport Streams array.
        //
        if (Wave->Streams[Channel] == this)
        {
            Wave->Streams[Channel] = NULL;
        }
        
        //
        // Release the scatter/gather table.
        //
        if (stPRDList.pPRDEntry)
        {
            HalFreeCommonBuffer (Wave->AdapterObject,
                                 (MAX_PRD_ENTRIES+1) * sizeof (tPRDEntryWin) * 2,
                                 stPRDList.PhysAddr,
                                 (PVOID)stPRDList.pPRDEntry,
                                 FALSE);
            stPRDList.pPRDEntry = NULL;
			stPRDList.pPRDEntryBackup = NULL;
        }

	    if (failSafeBuf)
		{
			HalFreeCommonBuffer(
				Wave->AdapterObject,
				FAILSAFE_BUF_SIZE,
        		failSafeBufPhys,
        		failSafeBuf,
				FALSE);

			failSafeBuf = NULL;
		}


        //
        // Release the miniport.
        //
        Wave->Release ();
        Wave = NULL;
    }

    //
    // Release the service group.
    //
    if (ServiceGroup)
    {
        ServiceGroup->Release ();
        ServiceGroup = NULL;
    }

    //
    // Release the mapping table.
    //
    if (stPRDList.pMapData)
    {
        ExFreePool (stPRDList.pMapData);
        stPRDList.pMapData = NULL;
		stPRDList.pMapDataBackup = NULL;
    }

    //
    // Release the port stream.
    //
    if (PortStream)
    {
        PortStream->Release ();
        PortStream = NULL;
    }
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::Init
 *****************************************************************************
 * This routine initializes the stream object, sets up the PRD, and programs
 * the PRD base address register for the pin being
 * initialized.
 */
NTSTATUS CMiniportWaveGEODEStream::Init
(
    IN  CMiniportWaveGEODE        *Miniport_,
    IN  PPORTWAVEPCISTREAM      PortStream_,
    IN  ULONG                   Channel_,
    IN  BOOLEAN                 Capture_,
    IN  PKSDATAFORMAT           DataFormat_,
    OUT PSERVICEGROUP           *ServiceGroup_
)
{
    PAGED_CODE ();

	// protect DMA engine from codec accesses
   	KIRQL  OldIrql;

	int i;

    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::Init]"));

    ASSERT (Miniport_);
    ASSERT (PortStream_);
    ASSERT (DataFormat_);
    ASSERT (ServiceGroup_);

    //
    // The rule here is that we return when we fail without a cleanup.
    // The destructor will relase the allocated memory.
    //
    NTSTATUS ntStatus = STATUS_SUCCESS;

    //
    // Initialize PRD info.
    //
    stPRDList.pPRDEntry = NULL;
    stPRDList.pMapData = NULL;
    stPRDList.nHead = 0;
    stPRDList.nTail = 0;
    stPRDList.ulTagCounter = 0;
    stPRDList.nPRDEntries = 0;

	failSafeBuf = NULL;

    //
    // Save miniport pointer and addref it.
    //
    Wave = Miniport_;
    Wave->AddRef ();

    //
    // Save portstream interface pointer and addref it.
    //
    PortStream = PortStream_;
    PortStream->AddRef ();

    //
    // Save channel ID and capture flag.
    //
    Channel = Channel_;
    Capture = Capture_;

    //
    // Store the stream pointer, it is used by the ISR.
    //
    Wave->Streams[Channel] = this;

    //
    // Save data format and current sample rate.
    //
    DataFormat = (PKSDATAFORMAT_WAVEFORMATEX)DataFormat_;
    CurrentRate = DataFormat->WaveFormatEx.nSamplesPerSec;

    //
    // Initialize the PRD spinlock.
    //
    KeInitializeSpinLock (&MapLock);

    //
    // Create a service group (a DPC abstraction/helper) to help with 
    // interrupts.
    //
    ntStatus = PcNewServiceGroup (&ServiceGroup, NULL);
    if (!NT_SUCCESS (ntStatus))
    {
        DOUT (DBG_ERROR, ("Failed to create a service group!"));
        return ntStatus;
    }

    //
    // Pass the ServiceGroup pointer to portcls.
    //
    *ServiceGroup_ = ServiceGroup;
    ServiceGroup->AddRef ();

    // 
    // Setup the PRD 
    // Allocate 32 entries of 8 bytes (one PRD entry). We allocate two tables
    // because we need one table as a backup.
    // The pointer is aligned on a 8 byte boundary (that's what we need).
    //
    PDMAOPERATIONS DmaOperations;

    //
    // Query for the new interface "DmaOperations" and use the function provided there
    // in case the interface is offered.
    //
    if (!NT_SUCCESS (Wave->DmaChannel->QueryInterface (IID_IDmaOperations, (PVOID *)&DmaOperations)))
    {
        stPRDList.pPRDEntry = (tPRDEntryWin *)HalAllocateCommonBuffer (Wave->AdapterObject,
                                   (MAX_PRD_ENTRIES+1) * sizeof (tPRDEntryWin) * 2,
                                   &stPRDList.PhysAddr,
                                   FALSE);

		failSafeBuf =
			(PUCHAR)HalAllocateCommonBuffer(
				Wave->AdapterObject,
				FAILSAFE_BUF_SIZE,
				&failSafeBufPhys,
				FALSE);
    }
    else
    {
        stPRDList.pPRDEntry = (tPRDEntryWin *)DmaOperations->AllocateCommonBuffer (
                                    (MAX_PRD_ENTRIES+1) * sizeof (tPRDEntryWin) * 2,
                                    &stPRDList.PhysAddr,
                                    FALSE);

		failSafeBuf =
			(PUCHAR)DmaOperations->AllocateCommonBuffer (
				FAILSAFE_BUF_SIZE,
				&failSafeBufPhys,
				FALSE);

        DmaOperations->Release();
    }

    if (!stPRDList.pPRDEntry)
    {
        DOUT (DBG_ERROR, ("Failed HalAllocateCommonBuffer!"));
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    if (!failSafeBuf)
    {
        DOUT (DBG_ERROR, ("Failed HalAllocateCommonBuffer!"));
        return STATUS_INSUFFICIENT_RESOURCES;
    }

	DURAUDIO_SetPRDAddress (pGeode, Channel, stPRDList.PhysAddr.LowPart, (unsigned long) stPRDList.pPRDEntry);

	//
	//    Fill PRD with Zeroes
	//
	for (i=0; i<=MAX_PRD_ENTRIES; ++i)
	{
		stPRDList.pPRDEntry[i].dwPtrToPhyAddress=0;
		stPRDList.pPRDEntry[i].wLength=0;
		stPRDList.pPRDEntry[i].wFlags=0;
	}

	//
	// Put a JMP from the bottom of the PRD to the top 
	//
	int TempTail=MAX_PRD_ENTRIES;

	stPRDList.pPRDEntry[TempTail].dwPtrToPhyAddress = stPRDList.PhysAddr.u.LowPart;
	stPRDList.pPRDEntry[TempTail].wLength			= 0;
	stPRDList.pPRDEntry[TempTail].wFlags			= JMP;

    // calculate the (backup) pointer.
    stPRDList.pPRDEntryBackup =
		(tPRDEntryWin *)stPRDList.pPRDEntry + (MAX_PRD_ENTRIES + 1);

	ASSERT (stPRDList.pPRDEntry);
	ASSERT (stPRDList.pPRDEntryBackup);
    //
    // Allocate a buffer for the 32 possible mappings. We allocate two tables
    // because we need one table as a backup
    //
    stPRDList.pMapData = 
        (tMapData *)ExAllocatePoolWithTag (NonPagedPool, sizeof(tMapData) * 
                                    MAX_PRD_ENTRIES * 2, 'atad');

    if (!stPRDList.pMapData)
    {
        DOUT (DBG_ERROR, ("Failed to allocate the back up buffer!"));
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // calculate the (backup) pointer.
    stPRDList.pMapDataBackup = (tMapData *)stPRDList.pMapData + MAX_PRD_ENTRIES;
      
	ASSERT (stPRDList.pMapData);
	ASSERT (stPRDList.pMapDataBackup);
    //
    // Store the base address of this DMA engine.
    //
    if (Capture)
    {
		// Stream Direction
		m_PRDDir = WAPI_IN;
        DOUT (DBG_STREAM, ("Direction set to CAPTURE"));
    }
    else    // render
    {
        // Stream Direction
        m_PRDDir = WAPI_OUT;
        DOUT (DBG_STREAM, ("Direction set to PLAYBACK"));
    }

    //
    // Reset the DMA 
    //
    ResetDMA ();

    //
    // Reset the position pointers.
    //
    TotalBytesMapped   = 0;
    TotalBytesReleased = 0;

    //
    // Now set the requested sample rate. In case of a failure, the object
    // gets destroyed and releases all memory etc.
    //
    ntStatus = SetFormat (DataFormat_);
    if (!NT_SUCCESS (ntStatus))
    {
        DOUT (DBG_ERROR, ("Stream init SetFormat call failed!"));
        return ntStatus;
    }

    //
    // Initialize the device state.
    //
    m_PowerState = PowerDeviceD0;


    PPREFETCHOFFSET PreFetchOffset;
    //
    // Query for the new interface "PreFetchOffset" and use
    // function offered there in case the interface is offered.
    //
    if (NT_SUCCESS(PortStream->QueryInterface(IID_IPreFetchOffset, (PVOID *)&PreFetchOffset)))
    {
        // why don't we pad by 32 sample frames
        PreFetchOffset->SetPreFetchOffset(32 * (DataFormat->WaveFormatEx.nChannels * 2)); 
        PreFetchOffset->Release();
    }


    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::NonDelegatingQueryInterface
 *****************************************************************************
 * Obtains an interface.  This function works just like a COM QueryInterface
 * call and is used if the object is not being aggregated.
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::NonDelegatingQueryInterface
(
    IN  REFIID  Interface,
    OUT PVOID * Object
)
{
    PAGED_CODE ();

    ASSERT (Object);

    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::NonDelegatingQueryInterface]"));

    //
    // Convert for IID_IMiniportWavePciStream
    //
    if (IsEqualGUIDAligned (Interface, IID_IMiniportWavePciStream))
    {
        *Object = (PVOID)(PMINIPORTWAVEPCISTREAM)this;
    }
    //
    // Convert for IID_IServiceSink
    //
    else if (IsEqualGUIDAligned (Interface, IID_IServiceSink))
    {
        *Object = (PVOID)(PSERVICESINK)this;
    }
    //
    // Convert for IID_IDrmAudioStream
    //
    else if (IsEqualGUIDAligned (Interface, IID_IDrmAudioStream))
    {
        *Object = (PVOID)(PDRMAUDIOSTREAM)this;
    }
    //
    // Convert for IID_IUnknown
    //
    else if (IsEqualGUIDAligned (Interface, IID_IUnknown))
    {
        *Object = (PVOID)(PUNKNOWN)(PMINIPORTWAVEPCISTREAM)this;
    }
    else
    {
        *Object = NULL;
        return STATUS_INVALID_PARAMETER;
    }

    ((PUNKNOWN)*Object)->AddRef ();
    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::GetAllocatorFraming
 *****************************************************************************
 * Returns the framing requirements for this device.
 * That is sample size (for one sample) and preferred frame (buffer) size.
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::GetAllocatorFraming
(
    OUT PKSALLOCATOR_FRAMING AllocatorFraming 
)
{
    PAGED_CODE ();

    ULONG SampleSize;

    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::GetAllocatorFraming]"));

    //
    // Determine sample size in bytes.  Always number of 
    // channels * 2 (because 16-bit).
    //
    SampleSize = DataFormat->WaveFormatEx.nChannels * 2;

    //
    // Report the minimum requirements.
    //
    AllocatorFraming->RequirementsFlags =
        KSALLOCATOR_REQUIREMENTF_SYSTEM_MEMORY | 
        KSALLOCATOR_REQUIREMENTF_PREFERENCES_ONLY;
    AllocatorFraming->Frames = 8;

	//
    // Currently, arbitrarily selecting 10ms as the frame target size.
    //
    AllocatorFraming->FrameSize		= SampleSize * (DataFormat->WaveFormatEx.nSamplesPerSec / 100);
    AllocatorFraming->FileAlignment = FILE_LONG_ALIGNMENT;
    AllocatorFraming->PoolType		= NonPagedPool;

    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::SetFormat
 *****************************************************************************    
 * This routine tests for proper data format (calls wave miniport) and sets
 * or changes the stream data format.
 * To figure out if the codec supports the sample rate, we just program the
 * sample rate and read it back. If it matches we return happy, if not then
 * we restore the sample rate and return unhappy.
 * We fail this routine if we are currently running (playing or recording).
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::SetFormat
(
    IN  PKSDATAFORMAT   Format
)
{
    PAGED_CODE ();

    ASSERT (Format);

    ULONG   TempRate;

	// protect DMA engine from codec accesses
	KIRQL  OldIrql;

    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::SetFormat]"));
    
    //
    // Change sample rate when we are in the stop or pause states - not 
    // while running!
    //
    if (DMAEngineState == DMA_ENGINE_ON)
    {
        return STATUS_UNSUCCESSFUL;
    }

    //
    // Ensure format falls in proper range and is supported.
    //
    NTSTATUS ntStatus = Wave->TestDataFormat (Format, (WavePins)(Channel << 1));
    if (!NT_SUCCESS (ntStatus))
        return ntStatus;

    //
    // Retrieve wave format portion.
    //
    PWAVEFORMATEX waveFormat = (PWAVEFORMATEX)(Format + 1);
    
    //
    // Save current rate in this context.
    //
    TempRate = waveFormat->nSamplesPerSec;
    
    //
	//  Set the sample rate
    //

	DURAUDIO_SetCodecRate(pGeode, Channel, TempRate);
 
    //
    // print information and save the format information.
    //
    DataFormat = (PKSDATAFORMAT_WAVEFORMATEX)Format;
    CurrentRate = TempRate;
    
    return STATUS_SUCCESS;
}
    
/*****************************************************************************
 * CMiniportWaveGEODEStream::SetContentId
 *****************************************************************************
 * This routine gets called by drmk.sys to pass the content to the driver.
 * The driver has to enforce the rights passed.
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::SetContentId
(
    IN  ULONG       contentId,
    IN  PCDRMRIGHTS drmRights
)
{
    PAGED_CODE ();

    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::SetContentId]"));

    //
    // If "drmRights->DigitalOutputDisable" is set, we need to disable S/P-DIF.
    // Currently, we don't have knowledge about the S/P-DIF interface. However,
    // in case you expanded the driver with S/P-DIF features you need to disable
    // S/P-DIF or fail SetContentId. If you have HW that has S/P-DIF turned on
    // by default and you don't know how to turn off (or you cannot do that)
    // then you must fail SetContentId.
    //
    // In our case, we assume the codec has no S/P-DIF or disabled S/P-DIF by
    // default, so we can ignore the flag.
    //
    // Store the copyright flag. We have to disable PCM recording if it's set.
    //
    if (!Wave->AdapterCommon->GetMiniportTopology ())
    {
        DOUT (DBG_ERROR, ("Topology pointer not set!"));
        return STATUS_UNSUCCESSFUL;
    }
    else
    {
        Wave->AdapterCommon->GetMiniportTopology ()->
            SetCopyProtectFlag (drmRights->CopyProtect);
    }

    //
    // We assume that if we can enforce the rights, that the old content
    // will be destroyed. We don't need to store the content id since we
    // have only one playback channel, so we are finished here.
    //
    
    return STATUS_SUCCESS;
}

/*****************************************************************************
 * Non paged code begins here
 *****************************************************************************
 */

#pragma code_seg()
/*****************************************************************************
 * CMiniportWaveGEODEStream::PowerChangeNotify
 *****************************************************************************
 * This functions saves and maintains the stream state through power changes.
 */
NTSTATUS CMiniportWaveGEODEStream::PowerChangeNotify
(
    IN  POWER_STATE NewState
)
{
    PAGED_CODE ();

    KIRQL       OldIrql;
    NTSTATUS    ntStatus = STATUS_SUCCESS;
    
    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::PowerChangeNotify]"));
    
    //
    // We don't have to check the power state, that's already done by the wave
    // miniport.
    //

    DOUT (DBG_POWER, ("Changing state to D%d.",
                     (ULONG)NewState.DeviceState - (ULONG)PowerDeviceD0));
    
    switch (NewState.DeviceState)
    {
        case PowerDeviceD0:
            //
            // If we are coming from D2 or D3 we have to restore the registers cause
            // there might have been a power loss.
            //
            if ((m_PowerState == PowerDeviceD3) || (m_PowerState == PowerDeviceD2))
            {
                //
                // The scatter gather list is already arranged. A reset of the DMA
                // brings all pointers to the default state. From there we can start.
                //
                
                // Acquire the mapping spin lock
                KeAcquireSpinLock (&MapLock,&OldIrql);
       
		        ntStatus = ResetDMA ();

                // Release the mapping spin lock
                KeReleaseSpinLock (&MapLock,OldIrql);
            }
            break;

        case PowerDeviceD1:
        case PowerDeviceD2:
        case PowerDeviceD3:
            //
            // If we power down to D2 or D3 we might loose power, so we have to be
            // aware of the DMA engine resetting. In that case a play would start
            // with scatter gather entry 0 (the current index is read only).
            // We just rearrange the scatter gather list (like we do on
            // RevokeMappings) so that the current buffer which is played is at
            // entry 0.
            //
                
            // Acquire the mapping spin lock
            KeAcquireSpinLock (&MapLock,&OldIrql);

            // Disable interrupts and stop DMA just in case.
            ResetDMA ();

            // Get current index 
            int nCurrentIndex =
				(int) DURAUDIO_GetCurrentPRDIndex (pGeode, Channel);

            //
            // First move the PRD list to the beginning.
            //
            // In case the DMA engine was stopped, current index may point to an
            // empty PRD entry. When we start the DMA engine it would then play this
            // (undefined) entry, so we check here for that condition.
            //
            if ((nCurrentIndex == ((stPRDList.nHead - 1) & PRD_MASK)) && (stPRDList.nPRDEntries != MAX_PRD_ENTRIES - 1))
            {
                nCurrentIndex = stPRDList.nHead;     // point to head
            }
    
            //
            // Move PRD list to (0-((current - head) & mask)) & mask, where
            // ((current - head) & mask) is the difference between head and
            // current index, no matter where they are :)
            //
            MovePRDList (stPRDList.nHead, (stPRDList.nTail - 1 ) & PRD_MASK, (0 - ((nCurrentIndex - stPRDList.nHead) & PRD_MASK)) & PRD_MASK);
        
            //
            // Update structure.
            //
            stPRDList.nHead = (0 - ((nCurrentIndex - stPRDList.nHead) & PRD_MASK)) & PRD_MASK;
            stPRDList.nTail = (stPRDList.nHead + stPRDList.nPRDEntries) & PRD_MASK;
            
            // release the mapping spin lock
            KeReleaseSpinLock (&MapLock,OldIrql);
            break;
    }
    
    //
    // Save the new state.  This local value is used to determine when to 
    // cache property accesses and when to permit the driver from accessing 
    // the hardware.
    //
    m_PowerState = NewState.DeviceState;
    DOUT (DBG_POWER, ("Entering D%d",(ULONG)m_PowerState - (ULONG)PowerDeviceD0));

    return ntStatus;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::MovePRDList
 *****************************************************************************
 * Moves the PRDList.
 * This function is used to remove entries from the scatter gather list or to
 * move the valid entries to the top. The mapping table which is hard linked
 * to the scatter gather entries is moved too.
 * The function does not change any variables in tPRDList.
 * The mapping spin lock must be held when calling this routine.
 * We use this function to remove mappings (RevokeMappings) or to rearrange the
 * list for powerdown/up management (the DMA starts at position zero again).
 * Note that there is a simple way of doing this also. When you zero the buffer
 * length in the scatter gather, the DMA engine ignores the entry and continues
 * with the next. But our way is more generic and if you ever want to port the
 * driver to another DMA engine you might be thankful for this code.
 */
void CMiniportWaveGEODEStream::MovePRDList
(
    IN  int nFirst, 
    IN  int nLast, 
    IN  int nNewPos
)
{
    DOUT (DBG_DMA, ("[CMiniportWaveGEODEStream::MovePRDList]"));
    
    //
    // Print information about the scatter gather list.
    //
    DOUT (DBG_DMA, ("Moving PRD entry %d-%d to %d.", nFirst, nLast, nNewPos));

    //
    // First copy the tables to a save place.
    //
    RtlCopyMemory ((PVOID)stPRDList.pPRDEntryBackup,
                   (PVOID)stPRDList.pPRDEntry,
                   sizeof (tPRDEntryWin) * MAX_PRD_ENTRIES);
    RtlCopyMemory ((PVOID)stPRDList.pMapDataBackup,
                   (PVOID)stPRDList.pMapData,
                   sizeof (tMapData) * MAX_PRD_ENTRIES);

    //
    // We move all the entries in blocks to the new position.
    //
    int nBlockCounter = 0;
    do
    {
        nBlockCounter++;
        //
        // We must copy the block when the index wraps around (ring buffer)
        // or we are at the last entry.
        //
        if (((nNewPos + nBlockCounter) == MAX_PRD_ENTRIES) ||   // wrap around
            ((nFirst + nBlockCounter) == MAX_PRD_ENTRIES) ||    // wrap around
            ((nFirst + nBlockCounter) == (nLast + 1)))          // last entry
        {
            //
            // copy one block (multiple entries).
            //
            RtlCopyMemory ((PVOID)&stPRDList.pPRDEntry[nNewPos],
                           (PVOID)&stPRDList.pPRDEntryBackup[nFirst],
                           sizeof (tPRDEntryWin) * nBlockCounter);
            RtlCopyMemory ((PVOID)&stPRDList.pMapData[nNewPos],
                           (PVOID)&stPRDList.pMapDataBackup[nFirst],
                           sizeof (tMapData) * nBlockCounter);
            // adjust the index
            nNewPos = (nNewPos + nBlockCounter) & PRD_MASK;
            nFirst = (nFirst + nBlockCounter) & PRD_MASK;
            nBlockCounter = 0;
        }
    // nBlockCounter should be zero when the end condition hits.
    } while (((nFirst + nBlockCounter - 1) & PRD_MASK) != nLast);
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::Service
 *****************************************************************************
 * This routine is called by the port driver in response to the interrupt
 * service routine requesting service on the stream's service group.
 * Requesting service on the service group results in a DPC being scheduled
 * that calls this routine when it runs.
 */
STDMETHODIMP_(void) CMiniportWaveGEODEStream::Service (void)
{
    DOUT (DBG_DMA, ("Service"));

    // release all mappings
    ReleaseUsedMappings ();

    // get new mappings
    GetNewMappings ();
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::NormalizePhysicalPosition
 *****************************************************************************
 * Given a physical position based on the actual number of bytes transferred,
 * this function converts the position to a time-based value of 100ns units.
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::NormalizePhysicalPosition
(
    IN OUT PLONGLONG PhysicalPosition
)
{
    ULONG SampleSize;

    DOUT (DBG_FUNC, ("NormalizePhysicalPosition"));

    //
    // Determine the sample size in bytes
    //
    SampleSize = DataFormat->WaveFormatEx.nChannels * 2;

    //
    // Calculate the time in 100ns steps.
    //
    *PhysicalPosition = (_100NS_UNITS_PER_SECOND / SampleSize * 
                         *PhysicalPosition) / CurrentRate;

    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::GetPosition
 *****************************************************************************
 * Gets the stream position. This is a byte count of the current position of
 * a stream running on a particular DMA engine.  We must return a sample
 * accurate count or the WaveDrv32 wave drift tests (35.2 & 36.2) will fail.
 *
 * The position is the sum of three parts:
 *     1) The total number of bytes in released buffers
 *     2) The position in the current buffer.
 *     3) The total number of bytes in played but not yet released buffers
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::GetPosition
(
    OUT PULONGLONG   Position
)
{
    KIRQL   OldIrql;
    ULONG   nCurrentIndex = 0;
    ULONG   PositionInCurrentBuffer = 0;

    ASSERT (Position);
               
    //
    // Acquire the mapping spin lock.
    //
    KeAcquireSpinLock (&MapLock, &OldIrql);

    //
    // Start with TotalBytesReleased (mappings released).
    //
    *Position = TotalBytesReleased;
    
    //
    // If we have entries in the list, we may have buffers that have not been
    // released but have been at least partially played.
    //
    if (stPRDList.nPRDEntries)
    {
        //
        // Repeat this until we get the same reading twice.  This will prevent
        // jumps when we are near the end of the buffer.
        //
        do
        {
            nCurrentIndex = DURAUDIO_GetCurrentPRDIndex(pGeode,Channel);
            
            PositionInCurrentBuffer =
				(WORD)DURAUDIO_GetPositionInBuffer (
					pGeode, Channel, nCurrentIndex);
        }
		while (nCurrentIndex !=
			(int) DURAUDIO_GetCurrentPRDIndex(pGeode,Channel));

		if ((nCurrentIndex == stPRDList.nTail) &&
			(stPRDList.nPRDEntries != MAX_PRD_ENTRIES))
		{
            PositionInCurrentBuffer = 0;
		}
        
        //
        // Add in our position in the current buffer.  The read returns the 
        // amount left in the buffer.
        //
        *Position += PositionInCurrentBuffer;
        
        //
        // Total any buffers that have been played and not released.
        //
        int i = stPRDList.nHead;
        while (i != nCurrentIndex)
        {
            *Position += (ULONGLONG)stPRDList.pMapData[i].ulBufferLength;
            i = (i + 1) & PRD_MASK;
        }
    }

    //
    // Release the mapping spin lock.
    //
    KeReleaseSpinLock (&MapLock, OldIrql);


    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::RevokeMappings
 *****************************************************************************
 * This routine is used by the port to revoke mappings previously delivered 
 * to the miniport stream that have not yet been unmapped.  This would 
 * typically be called in response to an I/O cancellation request.
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::RevokeMappings
(
    IN  PVOID     FirstTag,
    IN  PVOID     LastTag,
    OUT PULONG    MappingsRevoked
)
{
    ASSERT (MappingsRevoked);

    KIRQL   OldIrql;
    ULONG   ulOldDMAEngineState;
    int     nCurrentIndex, nSearchIndex, nFirst, nLast, nNumMappings;

    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::RevokeMappings]"));

    //
    // print information about the scatter gather list.
    //
    DOUT (DBG_FUNC, ("revoke: Head %d, Tail %d, Tag counter %d, Entries %d.",
                   stPRDList.nHead, stPRDList.nTail, stPRDList.ulTagCounter, 
                   stPRDList.nPRDEntries));

    //
    // Start accessing the mappings.
    //
    KeAcquireSpinLock (&MapLock, &OldIrql);

    //
    // Save old DMA engine state.
    //
    ulOldDMAEngineState = DMAEngineState;
    
    //
    // First stop the DMA engine so it won't process the next buffer in the
    // scatter gather list which might be one of the revoked buffers.
    //
    PauseDMA ();

    // Get current index 
    nCurrentIndex = DURAUDIO_GetCurrentPRDIndex(pGeode, Channel);

    //
    // We always rearrange the scatter gather list. That means we reset the DMA
    // engine and move the PRD list so that the entry where the current index
    // pointed to is located at position 0.
    //
    ResetDMA ();

    //
    // Return immediately if we just have 1 entry in our PRD list.
    //
    if (!stPRDList.nPRDEntries || (stPRDList.nPRDEntries == 1))
    {
        *MappingsRevoked = stPRDList.nPRDEntries;
        stPRDList.nHead = stPRDList.nTail = stPRDList.nPRDEntries = 0;
        
        KeReleaseSpinLock (&MapLock, OldIrql);
        return STATUS_SUCCESS;
    }


    //
    // First move the PRD list to the beginning.  In case the DMA engine was 
    // stopped, current index may point to an empty PRD entry.
    //
    if ((nCurrentIndex == ((stPRDList.nHead - 1) & PRD_MASK)) &&
       (stPRDList.nPRDEntries != MAX_PRD_ENTRIES - 1))
    {
        nCurrentIndex = stPRDList.nHead;     // point to head
    }

    //
    // Move PRD list to (0-((current - head) & mask)) & mask
    // where ((current - head) & mask) is the difference between head and
    // current index, no matter where they are :)
    //
	// This will place the current entry at 0 in the table for restart
	//
    MovePRDList (stPRDList.nHead, (stPRDList.nTail - 1 ) & PRD_MASK,
            (0 - ((nCurrentIndex - stPRDList.nHead) & PRD_MASK)) & PRD_MASK);
    
    //
    // Update structure. #entries still the same at this point
	// Table moved so current is at 0...
    //
    stPRDList.nHead = (0 - ((nCurrentIndex - stPRDList.nHead) & PRD_MASK)) & PRD_MASK;
    stPRDList.nTail = (stPRDList.nHead + stPRDList.nPRDEntries) & PRD_MASK;

	if ((stPRDList.nPRDEntries < MAX_PRD_ENTRIES))
	{
		int nTail = stPRDList.nTail;
		stPRDList.pPRDEntry[nTail].dwPtrToPhyAddress =
			failSafeBufPhys.u.LowPart;
		stPRDList.pPRDEntry[nTail].wLength = FAILSAFE_BUF_SIZE;
		stPRDList.pPRDEntry[nTail].wFlags = EOT;
	}

    //
    // Then we have to search for the tags. If we wouldn't have to rearrange the
    // scatter gather list all the time, then we could use the tag as an index
    // to the array, but the only way to clear DMA caches is a reset which has
    // the side-effect that we have to rearrange the PRD list (see above).
    //
    // search for first...
    //
    nSearchIndex = stPRDList.nHead;
    do
    {
        if ((void *)ULongToPtr(stPRDList.pMapData[nSearchIndex].ulTag) == FirstTag)
            break;
        nSearchIndex = (nSearchIndex + 1) & PRD_MASK;
    } while (nSearchIndex != stPRDList.nTail);
    nFirst = nSearchIndex;
    
    //
    // Search for last...
    //
    nSearchIndex = stPRDList.nHead;
    do
    {
        if ((void *)ULongToPtr(stPRDList.pMapData[nSearchIndex].ulTag) == LastTag)
            break;
        nSearchIndex = (nSearchIndex + 1) & PRD_MASK;
    } while (nSearchIndex != stPRDList.nTail);
    nLast = nSearchIndex;


    //
    // Check search result.
    //
    if ((nFirst == stPRDList.nTail) || (nLast == stPRDList.nTail))
    {
        DOUT (DBG_ERROR, ("!!! Entry not found !!!"));
        
        //
        // restart DMA in case it was running
        //
        if ((ulOldDMAEngineState == DMA_ENGINE_ON) && stPRDList.nPRDEntries)
		{
          	//  ResumeDMA ();
			DURAUDIO_StartDMA (pGeode, Channel);
			DMAEngineState = DMA_ENGINE_ON;
		}
        *MappingsRevoked = 0;
        KeReleaseSpinLock (&MapLock, OldIrql);

        return STATUS_UNSUCCESSFUL;         // one of the tags not found
    }

    // Print the index numbers found.
    DOUT (DBG_DMA, ("revoke: Removing entries %d (%d) to %d (%d).", nFirst, FirstTag,
                  nLast, LastTag));

    //
    // Calculate the entries between the indizes.
    //
    if (nLast < nFirst)
    {
        nNumMappings = ((nLast + MAX_PRD_ENTRIES) - nFirst) + 1;
    }
    else
    {
        nNumMappings = (nLast - nFirst) + 1;
    }

    //
    // Print debug inormation.
    //
    DOUT (DBG_DMA, ("Found entries: %d-%d, %d entries.", nFirst, nLast, 
                   nNumMappings));
    
    //
    // Now remove the revoked buffers.  Move the PRD list and modify the 
    // status information.
    //
    if (nFirst < stPRDList.nTail)
    {
        //
        // In this case, both first and last are >= the current index (0)
        //
        if (nLast != ((stPRDList.nTail - 1) & PRD_MASK))
        {
            //
            // Not the last entry, so move the PRD list + mappings.
            //
            MovePRDList ((nLast + 1) & PRD_MASK, (stPRDList.nTail - 1) & PRD_MASK,
                        nFirst);
        }
        stPRDList.nTail = (stPRDList.nTail - nNumMappings) & PRD_MASK;
    }

    //
    // In this case, at least first is "<" than current index (0)
    //
    else    
    {
        //
        // Check for last.
        //
        if (nLast < stPRDList.nTail)
        {                 
            //
            // Last is ">=" current index and first is "<" current index (0).
            // Remove MAX_PRD_ENTRIES - first entries in front of current index.
            //
            if (nFirst != stPRDList.nHead)
            {
                //
                // Move from head towards current index.
                //
               MovePRDList (stPRDList.nHead, nFirst - 1, 
                           (stPRDList.nHead + (MAX_PRD_ENTRIES - nFirst)) & 
                           PRD_MASK);
            }

            //
            // Adjust head.
            //
            stPRDList.nHead = (stPRDList.nHead + (MAX_PRD_ENTRIES - nFirst)) & 
                PRD_MASK;

            //
            // Remove nLast entries from CIV to tail.
            //
            if (nLast != ((stPRDList.nTail - 1) & PRD_MASK))
            {
                //
                // Not the last entry, so move the PRD list + mappings.
                //
                MovePRDList (nLast + 1, (stPRDList.nTail - 1) & PRD_MASK, 0);
            }

            //
            // Adjust tail.
            //
            stPRDList.nTail = (stPRDList.nTail - (nLast + 1)) & PRD_MASK;
        }

        //
        // Last is "<" current index and first is "<" current index (0).
        //
        else
        {
            //
            // Remove nNumMappings entries in front of current index.
            //
            if (nFirst != stPRDList.nHead)
            {
                //
                // Move from head towards current index.
                //
                MovePRDList (stPRDList.nHead, nFirst - 1, 
                           (nLast - nNumMappings) + 1);
            }
            
            //
            // Adjust head.
            //
            stPRDList.nHead = (stPRDList.nHead + nNumMappings) & PRD_MASK;
        }
    }

    //
    // In all cases, reduce the number of mappings.
    //
    stPRDList.nPRDEntries -= nNumMappings;

	if ((stPRDList.nPRDEntries < MAX_PRD_ENTRIES))
	{
		int nTail = stPRDList.nTail;
		stPRDList.pPRDEntry[nTail].dwPtrToPhyAddress =
			failSafeBufPhys.u.LowPart;
		stPRDList.pPRDEntry[nTail].wLength = FAILSAFE_BUF_SIZE;
		stPRDList.pPRDEntry[nTail].wFlags = EOT;
	}

    //
    // Print debug information.
    //
    DOUT (DBG_DMA, ("revoke: Number of mappings is now %d, Head is %d, Tail is %d",
            stPRDList.nPRDEntries, stPRDList.nHead, stPRDList.nTail));

	//
    // Just un-pause the DMA engine if it was running before and there are 
    // still entries left and tail != 0.
    //
    if ((ulOldDMAEngineState == DMA_ENGINE_ON) && stPRDList.nPRDEntries 
       && stPRDList.nTail)
    {
        // ResumeDMA ();
		DURAUDIO_StartDMA (pGeode, Channel);
		DMAEngineState = DMA_ENGINE_ON;
    }
	else
	{
	    DOUT (DBG_DMA, ("revoke: !!NOT RESUMING entries=%d, Head=%d, Tail=%d",
            stPRDList.nPRDEntries, stPRDList.nHead, stPRDList.nTail));

	}

    //
    // Release the mapping spin lock and return the number of mappings we 
    // revoked.
    //
    KeReleaseSpinLock (&MapLock, OldIrql);
    *MappingsRevoked = nNumMappings;
    
    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::MappingAvailable
 *****************************************************************************
 * This routine is called by the port driver to notify the stream that there
 * are new mappings available.  Note that this is ONLY called after the stream
 * has previously had a GetMapping() call fail due to lack of available
 * mappings.
 */
STDMETHODIMP_(void) CMiniportWaveGEODEStream::MappingAvailable (void)
{
    KIRQL OldIrql;

    DOUT (DBG_DMA, ("MappingAvailable"));

    //
    // Release processed mappings.
    //
    ReleaseUsedMappings ();

    //
    // Process the new mappings.
    //
    GetNewMappings ();
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::GetNewMappings
 *****************************************************************************
 * This routine is called when new mappings are available from the port driver.
 * The routine places mappings into the input mapping queue. GEODE can handle up
 * to 32 entries (descriptors). We program the DMA registers if we have at least
 * one mapping in the queue. The mapping spin lock must be held when calling
 * this routine.
 */
NTSTATUS CMiniportWaveGEODEStream::GetNewMappings (void)
{
    KIRQL OldIrql;
    
    ULONG       Flags;
    NTSTATUS    ntStatus = STATUS_SUCCESS;
    int         nInsertMappings = 0;
    int         nTail, TempTail;                  // aut. variable
    int         firstTail;
    int         lastAdded;
	int			nCurrentIndex;
	BOOLEAN		inFailSafe = FALSE;

    DOUT (DBG_DMA, ("[GetNewMappings]"));

    // acquire the mapping spin lock
    KeAcquireSpinLock (&MapLock,&OldIrql);

    //
    // Print debug information ...
    //
    DOUT (DBG_DMA, ("Start getting mappings at %d.", stPRDList.nTail));

    firstTail = stPRDList.nTail;

    //
    // Get available mappings up to the MAX_PRD_ENTRIES that we can hold in the PRD.
    //
	// we could get a full set of tiny buffers as with getpos test
	// which could keep us going around and around

    while (stPRDList.nPRDEntries < (MAX_PRD_ENTRIES-1))
    {
        //
        // Get the information from the list.
        //
        ULONG               ulTag = stPRDList.ulTagCounter;
        ULONG               ulBufferLength;
        PHYSICAL_ADDRESS    PhysAddr;
        PVOID               VirtAddr;


        // Release the mapping spin lock
        KeReleaseSpinLock (&MapLock,OldIrql);

        //
        // Try to get the mapping from the port.
        // Here comes the problem: When calling GetMapping or ReleaseMapping we
        // cannot hold our spin lock. So we need to buffer the return values and
        // stick the information into the structure later when we have the spin
        // lock acquired.
        //
        ntStatus = PortStream->GetMapping ((PVOID)ULongToPtr(ulTag),
                                           (PPHYSICAL_ADDRESS)&PhysAddr,
                                           &VirtAddr,
                                           &ulBufferLength,
                                           &Flags);
    
        // Acquire the mapping spin lock
        KeAcquireSpinLock (&MapLock,&OldIrql);
        
        //
        // Quit this loop when we run out of mappings.
        //
        if (!NT_SUCCESS (ntStatus))
        {
		   DOUT (DBG_DMA, ("Exit getting mappings at %d, no more data", stPRDList.nTail));
           break;
        }

		// limit of 65532 for GeodeGX stereo audio and 65534 for mono

       	ASSERT (ulBufferLength < 0x10000);

		//
        // Save the mapping.
        //
        nTail = stPRDList.nTail;
        stPRDList.pMapData[nTail].ulTag			 = ulTag;
        stPRDList.pMapData[nTail].PhysAddr		 = PhysAddr;
        stPRDList.pMapData[nTail].pVirtAddr		 = VirtAddr;
        stPRDList.pMapData[nTail].ulBufferLength = ulBufferLength;

        //
        // Fill in the PRD entry with pointer to physical address and length.
        //
        stPRDList.pPRDEntry[nTail].dwPtrToPhyAddress = stPRDList.pMapData[nTail].PhysAddr.LowPart;
        stPRDList.pPRDEntry[nTail].wLength			 = (WORD)(stPRDList.pMapData[nTail].ulBufferLength);

		// need this on EVERY PRD to allow proper mapping flow
		stPRDList.pPRDEntry[nTail].wFlags = EOP; 

    	lastAdded = nTail;

		// check on every add in case multiple tiny buffers added
		if (DMAEngineState == DMA_ENGINE_ON)
		{
			nCurrentIndex = (int)DURAUDIO_GetCurrentPRDIndex (pGeode, Channel);
			if (nCurrentIndex == lastAdded)
			{
				// already into failsafe
				inFailSafe = TRUE;
			}
		}

        DOUT (DBG_DMA, ("Added mapping: %d.", nTail));

		DOUT (DBG_DMA,
			("%c %04X %04X %08X\n", ((Capture)?'C':'P'),
				stPRDList.pPRDEntry[nTail].wFlags, 
				stPRDList.pPRDEntry[nTail].wLength, 
				stPRDList.pPRDEntry[nTail].dwPtrToPhyAddress));
        //
        // Update tail and number of entries.
        //
        stPRDList.nTail = (stPRDList.nTail + 1) & PRD_MASK;
        stPRDList.nPRDEntries++;
        stPRDList.ulTagCounter++; 
        TotalBytesMapped += (ULONGLONG)ulBufferLength;
        nInsertMappings++;

		if ((stPRDList.nPRDEntries < (MAX_PRD_ENTRIES-1)))
		{
			nTail = stPRDList.nTail;
			stPRDList.pPRDEntry[nTail].dwPtrToPhyAddress =
				failSafeBufPhys.u.LowPart;
			stPRDList.pPRDEntry[nTail].wLength = FAILSAFE_BUF_SIZE;
			stPRDList.pPRDEntry[nTail].wFlags = EOT;
		}
    } // end of while

	if (stPRDList.nPRDEntries >= (MAX_PRD_ENTRIES-1)) 
	{
	   DOUT (DBG_DMA, ("!!!Exit getting mappings at %d, out of buffer space.", stPRDList.nTail));
	}

    //
    // If there were processed mappings, program the DMA registers.
    //
    if (nInsertMappings)
    {
		if (DMAEngineState == DMA_ENGINE_STARVING)
		{

			KeStallExecutionProcessor(10);
			// engine has stopped
			// we need to reset prd pointer before restarting dma

			DURAUDIO_ResetPRDReg(
				pGeode,
				Channel, 
				pGeode->AudioChannel[Channel].
					PRD_AllocInfo.PhysicalAddress+(
					stPRDList.nHead*sizeof(PRD_ENTRY)));
			
			ResumeDMA();
		}
		else if (DMAEngineState == DMA_ENGINE_ON)
		{
			if (!(DURAUDIO_IsEnabled(pGeode, Channel)))
			{
				DURAUDIO_ResetPRDReg(
					pGeode,
					Channel, 
					pGeode->AudioChannel[Channel].
						PRD_AllocInfo.PhysicalAddress+(
							stPRDList.nHead*sizeof(PRD_ENTRY)));

				ResumeDMA();
			}
			else
			{
				if (inFailSafe)
				{
					// we've just crossed over into failsafe

					KeStallExecutionProcessor(10);

					DURAUDIO_ResetPRDReg(
						pGeode,
						Channel, 
						pGeode->AudioChannel[Channel].
							PRD_AllocInfo.PhysicalAddress+(
								stPRDList.nHead*sizeof(PRD_ENTRY)));

					ResumeDMA();
				}
			}
		}
    }

    //
    // Print debug information ... 
    //
   	DOUT (DBG_DMA, ("LG%c: %d %d\n", ((Capture)?'C':'P'),stPRDList.nHead, stPRDList.nTail));

    // Release the mapping spin lock
    KeReleaseSpinLock (&MapLock,OldIrql);
    
	return ntStatus;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::ReleaseUsedMappings
 *****************************************************************************
 * This routine unmaps previously mapped memory that the hardware has 
 * completed processing on.  This routine is typically called at DPC level 
 * from the stream deferred procedure call that results from a stream 
 * interrupt. The mapping spin lock must be held when calling this routine.
 */
NTSTATUS CMiniportWaveGEODEStream::ReleaseUsedMappings (void)
{
    KIRQL OldIrql;

    int nReleasedMappings = 0;

    DOUT (DBG_DMA, ("[CMiniportWaveGEODEStream::ReleaseUsedMappings]"));

    // acquire the mapping spin lock
    KeAcquireSpinLock (&MapLock,&OldIrql);

    //
    // Get current PRD index 
    //
	int nCurrentIndex = (int) DURAUDIO_GetCurrentPRDIndex (pGeode, Channel);

    //
    // Print debug info.
    //
    DOUT (DBG_DMA, ("Enter Release: Head %d, Tail %d, nPRDEntries:%d cur:%d.", stPRDList.nHead, stPRDList.nTail, stPRDList.nPRDEntries, nCurrentIndex));

    //
    // Return if no entries.
    //
    if (!stPRDList.nPRDEntries)
    {
        DOUT (DBG_DMA, ("No mappings to unmap."));
	    KeReleaseSpinLock (&MapLock,OldIrql);
        return STATUS_SUCCESS;
    }


    //
    // Clean up everything to that index.
    //
    while (stPRDList.nPRDEntries)
    {
    	int lastReleased;

        //
        // Check to see if we've released all the buffers.
        //

		// keep getting the current index inside this loop
		nCurrentIndex = (int) DURAUDIO_GetCurrentPRDIndex (pGeode, Channel);

		// this one (and only one) test is ok
		// need to check if HEAD has caught up to current position
		// which means we've released all buffers

        if (stPRDList.nHead == nCurrentIndex)
		{
			// only 1 block remaining and we're transferring it
			break;
		}

        //
        // Save the tag and remove the entry from the list.
        //
        ULONG   ulTag = stPRDList.pMapData[stPRDList.nHead].ulTag;
        TotalBytesReleased += (ULONGLONG)stPRDList.pMapData[stPRDList.nHead].ulBufferLength;

        stPRDList.nPRDEntries--;

		nReleasedMappings++;

        lastReleased = stPRDList.nHead;

        stPRDList.nHead = (stPRDList.nHead + 1) & PRD_MASK;

		// Due to lag in DMA engine clearing command enable bit, we need
		// to be pro-active in low data situations.  Setting pause when
		// no more buffers (eg still playing last buffer) is the best
		// solution.  It should keep us out of that race condition...

		if (stPRDList.nPRDEntries <= 0)
		{
			DMAEngineState = DMA_ENGINE_STARVING;
		}

		if ((nReleasedMappings == 1) &&
			(stPRDList.nPRDEntries == (MAX_PRD_ENTRIES-1)))
		{
			// we were full but now we have room to add failsafe

			int nTail = stPRDList.nTail;
			stPRDList.pPRDEntry[nTail].dwPtrToPhyAddress =
				failSafeBufPhys.u.LowPart;
			stPRDList.pPRDEntry[nTail].wLength = FAILSAFE_BUF_SIZE;
			stPRDList.pPRDEntry[nTail].wFlags = EOT;
		}

        // Release the mapping spin lock
        KeReleaseSpinLock (&MapLock,OldIrql);

        //
        // Release this entry.
        //
        PortStream->ReleaseMapping ((PVOID)ULongToPtr(ulTag));

        // acquire the mapping spin lock
        KeAcquireSpinLock (&MapLock,&OldIrql);

		DOUT (DBG_DMA, ("Released entry: %d, Tail:%d, TotalEntries: %d.", lastReleased, stPRDList.nTail, 									stPRDList.nPRDEntries));
    }

    //
    // Print release information and return.
    //

	DOUT (DBG_DMA, ("LR%c: %d %d %d", ((Capture)?'C':'P'),stPRDList.nHead, stPRDList.nTail, stPRDList.nPRDEntries));

    // Release the mapping spin lock
    KeReleaseSpinLock (&MapLock,OldIrql);


    return STATUS_SUCCESS;
}

/*****************************************************************************
 * CMiniportWaveGEODEStream::SetState
 *****************************************************************************
 * This routine sets/changes the DMA engine state to play, stop, or pause
 */
STDMETHODIMP_(NTSTATUS) CMiniportWaveGEODEStream::SetState
(
    IN  KSSTATE State
)
{
    PAGED_CODE ();      // Gets called at PASSIVE_LEVEL
    
    KIRQL OldIrql;
    
    DOUT (DBG_FUNC, ("[CMiniportWaveGEODEStream::SetState]"));
    DOUT (DBG_FUNC, ("SetState to %d", State));


    //
    // Start or stop the DMA engine dependent of the state.
    //
    switch (State)
    {
        case KSSTATE_STOP:

            // Just pause DMA. If we have mappings left in our queue,
            // portcls will call RevokeMappings to destroy them.

            // acquire the mapping spin lock
            KeAcquireSpinLock (&MapLock,&OldIrql);

			PauseDMA ();

            // release the mapping spin lock
            KeReleaseSpinLock (&MapLock,OldIrql);

            // Release processed mappings
            ReleaseUsedMappings ();

            // Reset the position counters.
            TotalBytesMapped = TotalBytesReleased = 0;
 
			DMAEngineState = DMA_ENGINE_OFF;


            break;
    
        case KSSTATE_ACQUIRE:
        case KSSTATE_PAUSE:
            // acquire the mapping spin lock
            KeAcquireSpinLock (&MapLock,&OldIrql);
            
            // pause now.
            PauseDMA ();

	        // release the mapping spin lock
	        KeReleaseSpinLock (&MapLock,OldIrql);        

            // Release processed mappings
            ReleaseUsedMappings ();

            break;
    
        case KSSTATE_RUN:
            //
            // Let's rock.
            //            
   
            // Make sure we are not running already.
            if (DMAEngineState == DMA_ENGINE_ON)
            {
            	KeAcquireSpinLock (&MapLock,&OldIrql);

				if (!(DURAUDIO_IsEnabled(pGeode, Channel)))		// if the dma engine shut itself off
			
				{
					// engine has stopped, we need to realign prd pointer and table, before restarting dma
					DURAUDIO_ResetPRDReg (pGeode, Channel, 
						(pGeode->AudioChannel[Channel].PRD_AllocInfo.PhysicalAddress+(stPRDList.nHead*sizeof(PRD_ENTRY))) ) ;
					ResumeDMA();
				}

            	KeReleaseSpinLock (&MapLock,OldIrql);

                return STATUS_SUCCESS;
            }

            // Release processed mappings.
            ReleaseUsedMappings ();

            // Get new mappings.
            GetNewMappings ();

			if (DMAEngineState == DMA_ENGINE_PAUSE || DMAEngineState == DMA_ENGINE_STARVING)
			{
				DOUT (DBG_DMA, ("Run: ResumeDMA"));
				// acquire the mapping spin lock
				KeAcquireSpinLock (&MapLock,&OldIrql);

				if (DMAEngineState == DMA_ENGINE_STARVING)
				{
					if (!(DURAUDIO_IsEnabled(pGeode, Channel)))
					{
						// DMA engine has stopped
						// we need to reset prd pointer before restarting dma

						DURAUDIO_ResetPRDReg(
							pGeode,
							Channel, 
							pGeode->AudioChannel[Channel].
								PRD_AllocInfo.PhysicalAddress+(stPRDList.nHead*sizeof(PRD_ENTRY)));
					}
				}

				DURAUDIO_ResumeDMA (pGeode, Channel);

				// Release the mapping spin lock.
				KeReleaseSpinLock (&MapLock,OldIrql);
			}
			else //Start running
			{
				DOUT (DBG_DMA, ("Run: StartDMA"));

				// acquire the mapping spin lock
				KeAcquireSpinLock (&MapLock,&OldIrql);

				DURAUDIO_StartDMA (pGeode, Channel);

				// Release the mapping spin lock.
				KeReleaseSpinLock (&MapLock,OldIrql);
			}

			DMAEngineState = DMA_ENGINE_ON;

            break;
	} //end of switch
    
    return STATUS_SUCCESS;
}

/*****************************************************************************
 * CMiniportWaveGEODEStream::ResetDMA
 *****************************************************************************
 * This routine resets the Run/Pause bit in the control register. In addition, it
 * resets all DMA registers contents.
 * To be safe, call it with the mappings spinlock.
 */
NTSTATUS CMiniportWaveGEODEStream::ResetDMA (void)
{
	DURAUDIO_StopDMA (pGeode, Channel);

    //
    // Set the DMA engine state.
    //
    DMAEngineState = DMA_ENGINE_OFF;


    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::PauseDMA
 *****************************************************************************
 * This routine pauses a hardware stream by reseting the Run/Pause bit in the
 * control registers, leaving DMA registers content intact so that the stream
 * can later be resumed.
 */
NTSTATUS CMiniportWaveGEODEStream::PauseDMA (void)
{
    //
    // Only pause if we're actually on.
    //

    if (DMAEngineState != DMA_ENGINE_ON)
    {
        return STATUS_SUCCESS;
    }

	DURAUDIO_PauseDMA (pGeode, Channel);

	//
	// Set the DMA engine state.
	//
	DMAEngineState = DMA_ENGINE_PAUSE;
    
    return STATUS_SUCCESS;
}


/*****************************************************************************
 * CMiniportWaveGEODEStream::ResumeDMA
 *****************************************************************************
 * This routine sets the Run/Pause bit for the particular DMA engine to resume
 * it after it's been paused. This assumes that DMA registers content have 
 * been preserved.
 */
NTSTATUS CMiniportWaveGEODEStream::ResumeDMA (void)
{
	DURAUDIO_ResumeDMA (pGeode, Channel);

    //
    // Set the DMA engine state.
    //
    DMAEngineState = DMA_ENGINE_ON;
    

    return STATUS_SUCCESS;
}

