/* 
 * <LIC_AMD_STD>
 * Copyright <2000-2010> ADVANCED MICRO DEVICES, INC. All Rights Reserved.
 *    
 *    
 * Unless otherwise designated in writing, this software and any related 
 * documentation are the confidential proprietary information of AMD. 
 *   
 * THESE MATERIALS ARE PROVIDED "AS IS" WITHOUT ANY UNLESS OTHERWISE 
 * NOTED IN WRITING, EXPRESS OR IMPLIED WARRANTY OF ANY KIND, INCLUDING 
 * BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, NONINFRINGEMENT, 
 * TITLE, FITNESS FOR ANY PARTICULAR PURPOSE AND IN NO EVENT SHALL AMD 
 * OR ITS LICENSORS BE LIABLE FOR ANY DAMAGES WHATSOEVER. 
 *   
 * AMD does not assume any responsibility for any errors which may appear 
 * in the Materials nor any responsibility to support or update the Materials.  
 * AMD retains the right to modify the Materials at any time, without notice, 
 * and is not obligated to provide such modified Materials to you. AMD is 
 * not obligated to furnish, support, or make any further information 
 * available to you.
 * 
 * </LIC_AMD_STD>  
 * <CTL_AMD_STD>
 * </CTL_AMD_STD>  
 * <DOC_AMD_STD>
 * </DOC_AMD_STD>  
 */

#ifndef _GEODEWAVE_H_
#define _GEODEWAVE_H_

#include "shared.h"

//*****************************************************************************
// Defines
//*****************************************************************************

//
// The scatter gather can (only) handle 32 entries
//
const int MAX_PRD_ENTRIES = 32;

//
// Mask for accessing the scatter gather entries with a counter.
//
const int PRD_MASK = MAX_PRD_ENTRIES-1;

//
// These defines reflect what the DMA is supposed to be.
//
const int DMA_ENGINE_OFF   = 0;
const int DMA_ENGINE_ON    = 1;
const int DMA_ENGINE_PAUSE = 2;
const int DMA_ENGINE_STARVING = 3;
 
#define FAILSAFE_BUF_SIZE 4 // one stereo sample

//*****************************************************************************
// Data Structures and Typedefs
//*****************************************************************************

//
// This is a description of one mapping entry. It contains information
// about the buffer and a tag which is used for "cancel mappings".
// For one mapping that is described here, there is an entry in the
// scatter gather engine.
//
typedef struct tagMapData
{
    ULONG               ulTag;                  //tag, a simple counter.
    PHYSICAL_ADDRESS    PhysAddr;               //phys. addr. of buffer
    PVOID               pVirtAddr;              //virt. addr. of buffer
    ULONG               ulBufferLength;         //buffer length
} tMapData;

//
// Structure to describe the GEODE Physical Descriptor Table (PRD).
// The entries are allocated at once in common memory (non-cached memory). 
// To avoid slow-down of CPU, the additional information for handling 
// this structure is stored in tPRDList.
//
typedef struct tagPRDEntryWin
{
    DWORD   dwPtrToPhyAddress;
    WORD    wLength;
    WORD    wFlags;
} tPRDEntryWin;

//
// Structure needed to keep track of the PRD tables.
// This structure is seperated from tPRDEntry because we want to
// have it in cached memory (and not along with tPRDEntry in non-
// cached memory).
//
typedef struct tagPRDList
{
    PHYSICAL_ADDRESS        PhysAddr;			// Physical address of PRDList
    volatile tPRDEntryWin   *pPRDEntry;			// Virtual Address of PRDList
    tPRDEntryWin            *pPRDEntryBackup;	// needed for rearranging the PRDList
    tMapData                *pMapData;			// mapping list
    tMapData                *pMapDataBackup;	// needed for rearranging the PRDList
    int                     nHead;				// index for the PRD Head
    int                     nTail;				// index for the PRD Tail
    unsigned long           ulTagCounter;		// the tag is a simple counter.
    int                     nPRDEntries;		// number of entries.
} tPRDList;




//*****************************************************************************
// Classes
//*****************************************************************************

/*****************************************************************************
 * CMiniportWaveGEODEStream
 *****************************************************************************
 * GEODE wave miniport stream.
 */
class CMiniportWaveGEODEStream : public IMiniportWavePciStream,
                               public IDrmAudioStream,
                               public CUnknown
{
private:
    //
    // CMiniportWaveGEODEStream private variables
    //
    CMiniportWaveGEODE *        Wave;					// Miniport Object
    ULONG                       Channel;				// channel this stream handles.
    BOOL                        Capture;				// TRUE=Capture,FALSE=Render
    ULONG                       CurrentRate;			// Current Sample Rate
    PSERVICEGROUP               ServiceGroup;			// service group helps with DPCs
    tPRDList                    stPRDList;				// needed for scatter gather org.

	PUCHAR						failSafeBuf;			// Additional buff for DMA overrun
	PHYSICAL_ADDRESS			failSafeBufPhys;		// Physical add. of the additional buff

    PPORTWAVEPCISTREAM          PortStream;				// Port Stream Interface
    PKSDATAFORMAT_WAVEFORMATEX  DataFormat;				// Data Format
    KSPIN_LOCK                  MapLock;				// for processing mappings.
    ULONG						m_ulPRDAddr;			// Offset of the stream's DMA registers.
    PUCHAR						m_pucPRDCmd;			// Pointer to the DMA Command register
    unsigned long				m_PRDDir;				// PRD Direction (Playback/Capture)
    ULONG						DMAEngineState;			// DMA engine state (STOP, PAUSE, RUN)
    ULONGLONG					TotalBytesMapped;		// factor in position calculation
    ULONGLONG					TotalBytesReleased;		// factor in position calculation
    DEVICE_POWER_STATE			m_PowerState;			// Current power state of the device.


    /*************************************************************************
     * CMiniportWaveGEODEStream methods
     *************************************************************************
     *
     * These are private member functions used internally by the object.  See
     * GEODEWAVE.CPP for specific descriptions.
     */

    //
    // Moves the PRD and associated mappings list around.
    //
    void MovePRDList
    (
        IN  int nFirst, 
        IN  int nLast, 
        IN  int nNewPos
    );

    //
    // Called when new mappings have to be processed.
    //
    NTSTATUS GetNewMappings (void);

    //
    // Called when we want to release some mappings.
    //
    NTSTATUS ReleaseUsedMappings (void);

    //
    // DMA start/stop/pause/reset routines.
    //
    NTSTATUS ResetDMA (void);
    NTSTATUS PauseDMA (void);
    NTSTATUS ResumeDMA (void);
    

public:
    /*************************************************************************
     * The following two macros are from STDUNK.H.  DECLARE_STD_UNKNOWN()
     * defines inline IUnknown implementations that use CUnknown's aggregation
     * support.  NonDelegatingQueryInterface() is declared, but it cannot be
     * implemented generically.  Its definition appears in GEODEWAVE.CPP.
     * DEFINE_STD_CONSTRUCTOR() defines inline a constructor which accepts
     * only the outer unknown, which is used for aggregation.  The standard
     * create macro (in GEODEWAVE.CPP) uses this constructor.
     */
    DECLARE_STD_UNKNOWN ();
    DEFINE_STD_CONSTRUCTOR (CMiniportWaveGEODEStream);

    ~CMiniportWaveGEODEStream ();
    
    /*************************************************************************
     * Include IMiniportWavePciStream (public/exported) methods.
     *************************************************************************
     */
    IMP_IMiniportWavePciStream;

    /*************************************************************************
     * Include IDrmAudioStream (public/exported) methods.
     *************************************************************************
     */
    IMP_IDrmAudioStream;
    
    /*************************************************************************
     * CMiniportWaveGEODEStream methods
     *************************************************************************
     */
    
    //
    // Initializes the Stream object.
    //
    NTSTATUS Init
    (
        IN  CMiniportWaveGEODE    *Miniport_,
        IN  PPORTWAVEPCISTREAM  PortStream,
        IN  ULONG               Channel,
        IN  BOOLEAN             Capture,
        IN  PKSDATAFORMAT       DataFormat,
        OUT PSERVICEGROUP *     ServiceGroup
    );

    //
    // This method is called when the device changes power states.
    //
    NTSTATUS PowerChangeNotify
    (
        IN  POWER_STATE NewState
    );

    //
    // Return the current sample rate.
    //
    ULONG GetCurrentSampleRate (void)
    {
        return CurrentRate;
    }

    //
    // Friends
    //
    friend
    NTSTATUS CMiniportWaveGEODE::InterruptServiceRoutine
    (
        IN  PINTERRUPTSYNC  InterruptSync,
        IN  PVOID           StaticContext
    );
};

#endif
